#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BITS 12                   /* Setting the number of bits to 12, 13*/
#define HASHING_SHIFT (BITS-8)    /* or 14 affects several constants.    */
#define MAX_VALUE (1 << BITS) - 1 /* Note that MS-DOS machines need to   */
#define MAX_CODE MAX_VALUE - 1    /* compile their code in large model if*/
                                  /* 14 bits are selected.               */
  #define TABLE_SIZE 5021         /* The string table size needs to be a */
								  /* prime number that is somewhat larger*/
                                  /* than 2**BITS.                       */

int *code_value;                  /* This is the code value array        */
unsigned int *prefix_code;        /* This array holds the prefix codes   */
unsigned char *append_character;  /* This array holds the appended chars */
unsigned char decode_stack[4000]; /* This array holds the decoded string */

/*
** This is the hashing routine.  It tries to find a match for the prefix+char
** string in the string table.  If it finds it, the index is returned.  If
** the string is not found, the first available index in the string table is
** returned instead.
*/

int find_match(unsigned int hash_prefix, unsigned int hash_character)
{
	int index;
	int offset;

	index = (hash_character << HASHING_SHIFT) ^ hash_prefix;
	
	if (index == 0)
		offset = 1;
	else
		offset = TABLE_SIZE - index;

	while (1)
	{
		if (code_value[index] == -1)
			return(index);

		if (prefix_code[index] == hash_prefix && append_character[index] == hash_character)
			return(index);

		index -= offset;

		if (index < 0)
			index += TABLE_SIZE;
	}
}

void output_code(unsigned char *output, unsigned int* outputOffset, unsigned int code, bool firstTime, bool flush)
{
	static int output_bit_count=0;
	static unsigned int output_bit_buffer=0;

	if (firstTime)
	{
		output_bit_count = 0;
		output_bit_buffer = 0;
	}

	output_bit_buffer |= (unsigned long) code << (32-BITS-output_bit_count);
	if (output_bit_count == 4 && flush)
		output_bit_count += 4;
	else
		output_bit_count += BITS;
	while (output_bit_count >= 8)
	{
		output[(*outputOffset)++] = output_bit_buffer >> 24;
		output_bit_buffer <<= 8;
		output_bit_count -= 8;
	}
}

/*
** This is the compression routine.  The code should be a fairly close
** match to the algorithm accompanying the article.
**
*/

#define NEXT_CODE 257

int compress(unsigned char* input, unsigned int inputSize, unsigned char* output)
{
	unsigned int next_code;
	unsigned int character;
	unsigned int string_code;
	unsigned int last_string_code;
	unsigned int index;
	unsigned int inputOffset = 0;
	unsigned int outputOffset = 0;
	bool firstTime = true;

	code_value       = new int[TABLE_SIZE];
	prefix_code      = new unsigned int[TABLE_SIZE];
	append_character = new unsigned char[TABLE_SIZE];

	next_code = NEXT_CODE;                 /* Next code is the next available string code*/

	for (int i = 0; i < TABLE_SIZE; i++)   /* Clear out the string table before starting */
		code_value[i] = -1;
	memset(prefix_code, 0, TABLE_SIZE * sizeof(unsigned int));
	memset(append_character, 0, TABLE_SIZE);

	string_code = input[inputOffset++];/* Get the first code                         */
	last_string_code = string_code;

	/*
	** This is the main loop where it all happens.  This loop runs util all of
	** the input has been exhausted.  Note that it stops adding codes to the
	** table after all of the possible codes have been defined.
	*/
	while (inputOffset < inputSize)
	{
		character = input[inputOffset++];

		index = find_match(string_code, character);/* See if the string is in */

		if (code_value[index] != -1)               /* the table.  If it is,   */
		{
			string_code = code_value[index];       /* get the code value.  If */
		}
		else                                       /* the string is not in the*/
		{                                          /* table, try to add it.   */
			if (next_code >= 0xfff)
			{
				output_code(output, &outputOffset, string_code, firstTime, false);  /* When a string is found  */
				firstTime = false;

				next_code = NEXT_CODE;

				for (int i = 0; i < TABLE_SIZE; i++)   /* Clear out the string table before starting */
					code_value[i] = -1;
				memset(prefix_code, 0, TABLE_SIZE * sizeof(unsigned int));
				memset(append_character, 0, TABLE_SIZE);

				string_code = character;
				continue;
			}

			code_value[index]       = next_code++;
			prefix_code[index]      = string_code;
			append_character[index] = character;

			output_code(output, &outputOffset, string_code, firstTime, false);  /* When a string is found  */
			firstTime = false;

			string_code  =character;          /* that is not in the table*/
		}                                     /* I output the last string*/
	}                                         /* after adding the new one*/

	/*
	** End of the main loop.
	*/
	printf("%03X\n", string_code);
	output_code(output, &outputOffset, string_code, false, false); /* Output the last code               */
	output_code(output, &outputOffset, 0, false, true);            /* This code flushes the output buffer*/

	// Unknown insertion
	output[outputOffset++] = 0;

	delete[] code_value;
	delete[] prefix_code;
	delete[] append_character;

	return outputOffset;
}
